//
//  Copyright (c) 2018 Artyom Beilis (Tonkikh)
//
//  Distributed under the Boost Software License, Version 1.0. (See
//  accompanying file LICENSE_1_0.txt or copy at
//  http://www.boost.org/LICENSE_1_0.txt)
//
#ifndef BOOST_NOWIDE_REPLACEMENT_HPP_INCLUDED
#define BOOST_NOWIDE_REPLACEMENT_HPP_INCLUDED

/// @file

/// \def BOOST_NOWIDE_REPLACEMENT_CHARACTER
/// Unicode character to be used to replace invalid UTF-8 sequences
#ifndef BOOST_NOWIDE_REPLACEMENT_CHARACTER
#define BOOST_NOWIDE_REPLACEMENT_CHARACTER 0xFFFD
#endif

#endif // boost/nowide/replacement.hpp
